import { BaseError, encodeAbiParameters, hashTypedData, parseAbiParameters, } from 'viem';
import { readContract, signTypedData as viemSignTypedData } from 'viem/actions';
import { getAction } from 'viem/utils';
import AGWAccountAbi from '../abis/AGWAccount.js';
import { EOA_VALIDATOR_ADDRESS, SESSION_KEY_VALIDATOR_ADDRESS, } from '../constants.js';
import { getAgwTypedSignature } from '../getAgwTypedSignature.js';
import { encodeSessionWithPeriodIds, getPeriodIdsForTransaction, } from '../sessions.js';
import { isEip712TypedData, transformEip712TypedData } from '../utils.js';
import { sendPrivySignTypedData } from './sendPrivyTransaction.js';
import { signEip712TransactionInternal, } from './signTransaction.js';
export async function signTypedData(client, signerClient, parameters, isPrivyCrossApp = false) {
    if (isPrivyCrossApp)
        return await sendPrivySignTypedData(client, parameters);
    // if the typed data is already a zkSync EIP712 transaction, don't try to transform it
    // to an AGW typed signature, just pass it through to the signer.
    if (isEip712TypedData(parameters)) {
        const rawSignature = await viemSignTypedData(signerClient, parameters);
        // Match the expect signature format of the AGW smart account so the result can be
        // directly used in eth_sendRawTransaction as the customSignature field
        const hookData = [];
        const validationHooks = await getAction(client, readContract, 'readContract')({
            address: client.account.address,
            abi: AGWAccountAbi,
            functionName: 'listHooks',
            args: [true],
        });
        for (const _ of validationHooks) {
            hookData.push('0x');
        }
        const signature = encodeAbiParameters(parseAbiParameters(['bytes', 'address', 'bytes[]']), [rawSignature, EOA_VALIDATOR_ADDRESS, hookData]);
        return signature;
    }
    return await getAgwTypedSignature({
        client,
        signer: signerClient,
        messageHash: hashTypedData(parameters),
    });
}
export async function signTypedDataForSession(client, signerClient, parameters, session, paymasterHandler) {
    // if the typed data is already a zkSync EIP712 transaction, don't try to transform it
    // to an AGW typed signature, just pass it through to the signer.
    if (!isEip712TypedData(parameters)) {
        throw new BaseError('Session client can only sign EIP712 transactions as typed data');
    }
    const transactionRequest = transformEip712TypedData(parameters);
    if (!transactionRequest.to) {
        throw new BaseError('Transaction must have a to address');
    }
    // Match the expect signature format of the AGW smart account so the result can be
    // directly used in eth_sendRawTransaction as the customSignature field
    const validationHookData = {
        [SESSION_KEY_VALIDATOR_ADDRESS]: encodeSessionWithPeriodIds(session, getPeriodIdsForTransaction({
            sessionConfig: session,
            target: transactionRequest.to,
            selector: (transactionRequest.data?.slice(0, 10) ?? '0x'),
            timestamp: BigInt(Math.floor(Date.now() / 1000)),
        })),
    };
    const { customSignature } = await signEip712TransactionInternal(client, signerClient, {
        chain: client.chain,
        ...transactionRequest,
    }, SESSION_KEY_VALIDATOR_ADDRESS, false, validationHookData, paymasterHandler);
    return customSignature;
}
//# sourceMappingURL=signTypedData.js.map